Index: include/asterisk/alaw.h
===================================================================
--- include/asterisk/alaw.h	(revision 78064)
+++ include/asterisk/alaw.h	(working copy)
@@ -23,21 +23,63 @@
 #ifndef _ASTERISK_ALAW_H
 #define _ASTERISK_ALAW_H
 
+/* #define G711_REDUCED_BRANCHING */
+
 /*! Init the ulaw conversion stuff */
 /*!
  * To init the ulaw to slinear conversion stuff, this needs to be run.
  */
 void ast_alaw_init(void);
 
-/*! converts signed linear to mulaw */
+#define AST_ALAW_BIT_LOSS  4
+#define AST_ALAW_STEP      (1 << AST_ALAW_BIT_LOSS)
+#define AST_ALAW_TAB_SIZE  (32768 / AST_ALAW_STEP + 1)
+#define AST_ALAW_SIGN_BIT  0x80
+#define AST_ALAW_AMI_MASK  0x55
+
+
+/*! converts signed linear to alaw */
 /*!
   */
-extern unsigned char __ast_lin2a[8192];
+extern unsigned char __ast_lin2a[AST_ALAW_TAB_SIZE];
 
 /*! help */
 extern short __ast_alaw[256];
 
-#define AST_LIN2A(a) (__ast_lin2a[((unsigned short)(a)) >> 3])
+#define AST_LIN2A_LOOKUP(mag) \
+		__ast_lin2a[(mag) >> AST_ALAW_BIT_LOSS]
+
+/*! convert signed linear sample to sign-magnitude pair for a-Law */
+static inline void ast_alaw_get_sign_mag(short sample, unsigned *sign, unsigned *mag)
+{
+#if defined(G711_REDUCED_BRANCHING)
+	unsigned dual_mag;
+
+	*sign = ((unsigned short)sample >> 8) & AST_ALAW_SIGN_BIT;
+	dual_mag = (-sample << 16) | (unsigned short)sample;
+	*mag = (dual_mag >> (*sign >> 3)) & 0xffffU;
+	*sign ^= AST_ALAW_SIGN_BIT;
+#else
+	if (sample < 0)
+	{
+		*mag = -sample;
+		*sign = 0;
+	}
+	else
+	{
+		*mag = sample;
+		*sign = AST_ALAW_SIGN_BIT;
+	}
+#endif /* G711_REDUCED_BRANCHING */
+}
+
+static inline unsigned char AST_LIN2A(short sample)
+{
+	unsigned mag, sign;
+	ast_alaw_get_sign_mag(sample, &sign, &mag);
+	return (sign | AST_LIN2A_LOOKUP(mag)) ^ AST_ALAW_AMI_MASK;
+}
+
 #define AST_ALAW(a) (__ast_alaw[(int)(a)])
 
 #endif /* _ASTERISK_ALAW_H */
Index: include/asterisk/ulaw.h
===================================================================
--- include/asterisk/ulaw.h	(revision 78064)
+++ include/asterisk/ulaw.h	(working copy)
@@ -23,21 +23,61 @@
 #ifndef _ASTERISK_ULAW_H
 #define _ASTERISK_ULAW_H
 
+/* #define G711_REDUCED_BRANCHING */
+
 /*! Init the ulaw conversion stuff */
 /*!
  * To init the ulaw to slinear conversion stuff, this needs to be run.
  */
 void ast_ulaw_init(void);
 
+#define AST_ULAW_BIT_LOSS  3
+#define AST_ULAW_STEP      (1 << AST_ULAW_BIT_LOSS)
+#define AST_ULAW_TAB_SIZE  (32768 / AST_ULAW_STEP + 1)
+#define AST_ULAW_SIGN_BIT  0x80
+
+
 /*! converts signed linear to mulaw */
 /*!
   */
-extern unsigned char __ast_lin2mu[16384];
+extern unsigned char __ast_lin2mu[AST_ULAW_TAB_SIZE];
 
 /*! help */
 extern short __ast_mulaw[256];
 
-#define AST_LIN2MU(a) (__ast_lin2mu[((unsigned short)(a)) >> 2])
+#define AST_LIN2MU_LOOKUP(mag) \
+		__ast_lin2mu[((mag) + AST_ULAW_STEP / 2) >> AST_ULAW_BIT_LOSS]
+
+/*! convert signed linear sample to sign-magnitude pair for u-Law */
+static inline void ast_ulaw_get_sign_mag(short sample, unsigned *sign, unsigned *mag)
+{
+#if defined(G711_REDUCED_BRANCHING)
+	unsigned dual_mag;
+
+	*sign = ((unsigned short)sample >> 8) & AST_ULAW_SIGN_BIT;
+	dual_mag = (-sample << 16) | (unsigned short)sample;
+	*mag = (dual_mag >> (*sign >> 3)) & 0xffffU;
+#else
+	if (sample < 0)
+	{
+		*mag = -sample;
+		*sign = AST_ULAW_SIGN_BIT;
+	}
+	else
+	{
+		*mag = sample;
+		*sign = 0;
+	}
+#endif /* G711_REDUCED_BRANCHING */
+}
+
+static inline unsigned char AST_LIN2MU(short sample)
+{
+	unsigned mag, sign;
+	ast_ulaw_get_sign_mag(sample, &sign, &mag);
+	return ~(sign | AST_LIN2MU_LOOKUP(mag));
+}
+
 #define AST_MULAW(a) (__ast_mulaw[(a)])
 
 #endif /* _ASTERISK_ULAW_H */
Index: main/Makefile
===================================================================
--- main/Makefile	(revision 78064)
+++ main/Makefile	(working copy)
@@ -32,6 +32,7 @@
 # otherwise modules will not have them available if none of the static
 # objects use it.
 OBJS+=stdtime/localtime.o
+ASTCFLAGS+=-DTEST_TANDEM_TRANSCODING -DTEST_CODING_TABLES
 
 # At the moment say.o is an optional component which can be overridden
 # by a module.
Index: main/alaw.c
===================================================================
--- main/alaw.c	(revision 78064)
+++ main/alaw.c	(working copy)
@@ -18,84 +18,139 @@
 
 /*! \file
  *
- * \brief u-Law to Signed linear conversion
+ * \brief a-Law to Signed linear conversion
  *
  * \author Mark Spencer <markster@digium.com> 
  */
 
+/* significant enhancement dated 3 Aug 2007:
+This version may look slower than the original blind table lookups, 
+and it probably is slower on some systems indeed, but it is correct. 
+This version also shrinks the linear->law tables to 1/4 of their 
+original size, which puts a lot less strain on CPU's level 1 cache, 
+so it is probably faster on some systems too (I did not run any real benchmarks).
+
+You can enable the #define G711_REDUCED_BRANCHING for systems with 
+expensive branching (put it in ASTCFLAGS if you like), but when building 
+with -march=i686 it is not necessary because of 
+gcc's use of wonderful CMOV opcodes. */
+
 #include "asterisk.h"
 
 ASTERISK_FILE_VERSION(__FILE__, "$Revision$")
 
 #include "asterisk/alaw.h"
+#include "asterisk/logger.h"
 
-#define AMI_MASK 0x55
-
-static inline unsigned char linear2alaw (short int linear)
+static inline unsigned char linear2alaw(short sample, int full_coding)
 {
-    int mask;
-    int seg;
-    int pcm_val;
-    static int seg_end[8] =
-    {
-         0xFF, 0x1FF, 0x3FF, 0x7FF, 0xFFF, 0x1FFF, 0x3FFF, 0x7FFF
-    };
-    
-    pcm_val = linear;
-    if (pcm_val >= 0)
-    {
-        /* Sign (7th) bit = 1 */
-        mask = AMI_MASK | 0x80;
-    }
-    else
-    {
-        /* Sign bit = 0 */
-        mask = AMI_MASK;
-        pcm_val = -pcm_val;
-    }
-
-    /* Convert the scaled magnitude to segment number. */
-    for (seg = 0;  seg < 8;  seg++)
-    {
-        if (pcm_val <= seg_end[seg])
-	    break;
-    }
-    /* Combine the sign, segment, and quantization bits. */
-    return  ((seg << 4) | ((pcm_val >> ((seg)  ?  (seg + 3)  :  4)) & 0x0F)) ^ mask;
+	static const unsigned exp_lut[128] = {
+		1,1,2,2,3,3,3,3,
+		4,4,4,4,4,4,4,4,
+		5,5,5,5,5,5,5,5,
+		5,5,5,5,5,5,5,5,
+		6,6,6,6,6,6,6,6,
+		6,6,6,6,6,6,6,6,
+		6,6,6,6,6,6,6,6,
+		6,6,6,6,6,6,6,6,
+		7,7,7,7,7,7,7,7,
+		7,7,7,7,7,7,7,7,
+		7,7,7,7,7,7,7,7,
+		7,7,7,7,7,7,7,7,
+		7,7,7,7,7,7,7,7,
+		7,7,7,7,7,7,7,7,
+		7,7,7,7,7,7,7,7,
+		7,7,7,7,7,7,7,7 };
+	unsigned sign, exponent, mantissa, mag;
+	unsigned char alawbyte;
+	
+	ast_alaw_get_sign_mag(sample, &sign, &mag);
+	if (mag > 32767)
+		mag = 32767;        /* clip the magnitude for -32768 */
+	
+	exponent = exp_lut[(mag >> 8) & 0x7f];
+	mantissa = (mag >> (exponent + 3)) & 0x0f;
+	if (mag < 0x100)
+		exponent = 0;
+	
+	if (full_coding) {
+		/* full encoding, with sign and xform */
+		alawbyte = (unsigned char)(sign | (exponent << 4) | mantissa);
+		alawbyte ^= AST_ALAW_AMI_MASK;
+	} else {
+		/* half-cooked coding -- mantissa+exponent only (for lookup tab) */
+		alawbyte = (exponent << 4) | mantissa;
+	}
+	
+	return alawbyte;
 }
-/*- End of function --------------------------------------------------------*/
 
-static inline short int alaw2linear (unsigned char alaw)
+static inline short alaw2linear(unsigned char alawbyte)
 {
-    int i;
-    int seg;
-
-    alaw ^= AMI_MASK;
-    i = ((alaw & 0x0F) << 4);
-    seg = (((int) alaw & 0x70) >> 4);
-    if (seg)
-        i = (i + 0x100) << (seg - 1);
-    return (short int) ((alaw & 0x80)  ?  i  :  -i);
+	unsigned exponent, mantissa;
+	short sample;
+	
+	alawbyte ^= AST_ALAW_AMI_MASK;
+	exponent = (alawbyte & 0x70) >> 4;
+	mantissa = alawbyte & 0x0f;
+	sample = (mantissa << 4) + 8 /* rounding error */;
+	if (exponent)
+		sample = (sample + 0x100) << (exponent - 1);
+	if (!(alawbyte & 0x80))
+		sample = -sample;
+	return sample;
 }
 
-unsigned char __ast_lin2a[8192];
+unsigned char __ast_lin2a[AST_ALAW_TAB_SIZE];
 short __ast_alaw[256];
 
 void ast_alaw_init(void)
 {
 	int i;
+	
 	/* 
-	 *  Set up mu-law conversion table
+	 *  Set up a-law conversion table
 	 */
-	for(i = 0;i < 256;i++)
-	   {
-	        __ast_alaw[i] = alaw2linear(i);
-	   }
-	  /* set up the reverse (mu-law) conversion table */
-	for(i = -32768; i < 32768; i++)
-	   {
-		__ast_lin2a[((unsigned short)i) >> 3] = linear2alaw(i);
-	   }
-
+	for(i = 0; i < 256; i++) {
+		__ast_alaw[i] = alaw2linear(i);
+	}
+	/* set up the reverse (a-law) conversion table */
+	for(i = 0; i <= 32768; i += AST_ALAW_STEP) {
+		AST_LIN2A_LOOKUP(i) = linear2alaw(i, 0 /* half-cooked */);
+	}
+	
+#ifdef TEST_CODING_TABLES
+	for (i = -32768; i < 32768; ++i) {
+		unsigned char e1 = linear2alaw(i, 1);
+		short d1 = alaw2linear(e1);
+		unsigned char e2 = AST_LIN2A(i);
+		short d2 = alaw2linear(e2);
+		short d3 = AST_ALAW(e1);
+		
+		if (e1 != e2 || d1 != d3 || d2 != d3) {
+			ast_log(LOG_WARNING, "a-Law coding tables test failed on %d: e1=%u, e2=%u, d1=%d, d2=%d\n",
+					i, (unsigned)e1, (unsigned)e2, (int)d1, (int)d2);
+		}
+	}
+	ast_log(LOG_WARNING, "a-Law coding tables test completed.\n");
+#endif /* TEST_CODING_TABLES */
+	
+#ifdef TEST_TANDEM_TRANSCODING
+	/* tandem transcoding test */
+	for (i = -32768; i < 32768; ++i) {
+		unsigned char e1 = AST_LIN2A(i);
+		short d1 = AST_ALAW(e1);
+		unsigned char e2 = AST_LIN2A(d1);
+		short d2 = AST_ALAW(e2);
+		unsigned char e3 = AST_LIN2A(d2);
+		short d3 = AST_ALAW(e3);
+		
+		if (e1 != e2 || e2 != e3 || d1 != d2 || d2 != d3) {
+			ast_log(LOG_WARNING, "a-Law tandem transcoding test failed on %d: e1=%u, e2=%u, d1=%d, d2=%d, d3=%d\n",
+					i, (unsigned)e1, (unsigned)e2, (int)d1, (int)d2, (int)d3);
+		}
+	}
+	ast_log(LOG_WARNING, "a-Law tandem transcoding test completed.\n");
+#endif /* TEST_TANDEM_TRANSCODING */
 }
 
Index: main/ulaw.c
===================================================================
--- main/ulaw.c	(revision 78064)
+++ main/ulaw.c	(working copy)
@@ -23,23 +23,44 @@
  * \author Mark Spencer <markster@digium.com> 
  */
 
+/* significant enhancement dated 3 Aug 2007:
+This version may look slower than the original blind table lookups,
+and it probably is slower on some systems indeed, but it is correct.
+This version also shrinks the linear->law tables to 1/4 of their
+original size, which puts a lot less strain on CPU's level 1 cache,
+so it is probably faster on some systems too (I did not run any real benchmarks).
+
+You can enable the #define G711_REDUCED_BRANCHING for systems with
+expensive branching (put it in ASTCFLAGS if you like), but when building
+with -march=i686 it is not necessary because of
+gcc's use of wonderful CMOV opcodes. */
+
 #include "asterisk.h"
 
 ASTERISK_FILE_VERSION(__FILE__, "$Revision$")
 
 #include "asterisk/ulaw.h"
+#include "asterisk/logger.h"
 
+#if 0
+/* ZEROTRAP is the military recommendation to improve the encryption
+ * of u-Law traffic. It is irrelevant with modern encryption systems
+ * like AES, and will simply degrade the signal quality.
+ * ZEROTRAP is not implemented in AST_LIN2MU and so the coding table
+ * tests will fail if you use it */
 #define ZEROTRAP    /*!< turn on the trap as per the MIL-STD */
+#endif
+
 #define BIAS 0x84   /*!< define the add-in bias for 16 bit samples */
 #define CLIP 32635
 
-unsigned char __ast_lin2mu[16384];
+unsigned char __ast_lin2mu[AST_ULAW_TAB_SIZE];
 short __ast_mulaw[256];
 
 
-static unsigned char linear2ulaw(short sample)
+static unsigned char linear2ulaw(short sample, int full_coding)
 {
-	static int exp_lut[256] = {
+	static const unsigned exp_lut[256] = {
 		0,0,1,1,2,2,2,2,3,3,3,3,3,3,3,3,
 		4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,
 		5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
@@ -56,51 +77,103 @@
 		7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
 		7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
 		7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7 };
-	int sign, exponent, mantissa;
+	unsigned sign, exponent, mantissa, mag;
 	unsigned char ulawbyte;
-
+	
 	/* Get the sample into sign-magnitude. */
-	sign = (sample >> 8) & 0x80;          /* set aside the sign */
-	if (sign != 0) 
-		sample = -sample;              /* get magnitude */
-	if (sample > CLIP)
-		sample = CLIP;             /* clip the magnitude */
-
+	ast_ulaw_get_sign_mag(sample, &sign, &mag);
+	if (mag > CLIP)
+		mag = CLIP;             /* clip the magnitude */
+	
 	/* Convert from 16 bit linear to ulaw. */
-	sample = sample + BIAS;
-	exponent = exp_lut[(sample >> 7) & 0xFF];
-	mantissa = (sample >> (exponent + 3)) & 0x0F;
-	ulawbyte = ~(sign | (exponent << 4) | mantissa);
+	mag += BIAS;
+	exponent = exp_lut[(mag >> 7) & 0xFF];
+	mantissa = (mag >> (exponent + 3)) & 0x0F;
+	
+	if (full_coding) {
+		/* full encoding, with sign and xform */
+		ulawbyte = ~(sign | (exponent << 4) | mantissa);
 #ifdef ZEROTRAP
-	if (ulawbyte == 0)
-		ulawbyte = 0x02;   /* optional CCITT trap */
+		if (ulawbyte == 0)
+			ulawbyte = 0x02;   /* optional CCITT trap */
 #endif
-
+	} else {
+		/* half-cooked coding -- mantissa+exponent only (for lookup tab) */
+		ulawbyte = (exponent << 4) | mantissa;
+	}
+	
 	return ulawbyte;
 }
 
+static inline short ulaw2linear(unsigned char ulawbyte)
+{
+	unsigned exponent, mantissa;
+	short sample;
+	static const short etab[]={0,132,396,924,1980,4092,8316,16764};
+
+	ulawbyte = ~ulawbyte;
+	exponent = (ulawbyte & 0x70) >> 4;
+	mantissa = ulawbyte & 0x0f;
+	sample = mantissa << (exponent + 3);
+	sample += etab[exponent];
+	if (ulawbyte & 0x80)
+		sample = -sample;
+	return sample;
+}
+
 /*!
  * \brief  Set up mu-law conversion table
  */
 void ast_ulaw_init(void)
 {
 	int i;
-	for (i = 0; i < 256; i++) {
-		short mu, e, f, y;
-		static short etab[] = {0,132,396,924,1980,4092,8316,16764};
 
-		mu = 255 - i;
-		e = (mu & 0x70) / 16;
-		f = mu & 0x0f;
-		y = f * (1 << (e + 3));
-		y += etab[e];
-		if (mu & 0x80)
-			y = -y;
-		__ast_mulaw[i] = y;
+	/* 
+	 *  Set up mu-law conversion table
+	 */
+	for(i = 0; i < 256; i++) {
+		__ast_mulaw[i] = ulaw2linear(i);
 	}
 	/* set up the reverse (mu-law) conversion table */
-	for (i = -32768; i < 32768; i++) {
-		__ast_lin2mu[((unsigned short)i) >> 2] = linear2ulaw(i);
+	for(i = 0; i <= 32768; i += AST_ULAW_STEP) {
+		AST_LIN2MU_LOOKUP(i) = linear2ulaw(i, 0 /* half-cooked */);
 	}
+
+#ifdef TEST_CODING_TABLES
+	for (i = -32768; i < 32768; ++i) {
+		unsigned char e1 = linear2ulaw(i, 1);
+		short d1 = ulaw2linear(e1);
+		unsigned char e2 = AST_LIN2MU(i);
+		short d2 = ulaw2linear(e2);
+		short d3 = AST_MULAW(e1);
+		
+		if (e1 != e2 || d1 != d3 || d2 != d3) {
+			ast_log(LOG_WARNING, "u-Law coding tables test failed on %d: e1=%u, e2=%u, d1=%d, d2=%d\n",
+					i, (unsigned)e1, (unsigned)e2, (int)d1, (int)d2);
+		}
+	}
+	ast_log(LOG_WARNING, "u-Law coding tables test completed.\n");
+#endif /* TEST_CODING_TABLES */
+
+#ifdef TEST_TANDEM_TRANSCODING
+ 	/* tandem transcoding test */
+	for (i = -32768; i < 32768; ++i) {
+		unsigned char e1 = AST_LIN2MU(i);
+		short d1 = AST_MULAW(e1);
+		unsigned char e2 = AST_LIN2MU(d1);
+		short d2 = AST_MULAW(e2);
+		unsigned char e3 = AST_LIN2MU(d2);
+		short d3 = AST_MULAW(e3);
+		
+		if (i < 0 && e1 == 0x7f && e2 == 0xff && e3 == 0xff)
+			continue; /* known and normal negative 0 case */
+
+		if (e1 != e2 || e2 != e3 || d1 != d2 || d2 != d3) {
+			ast_log(LOG_WARNING, "u-Law tandem transcoding test failed on %d: e1=%u, e2=%u, d1=%d, d2=%d, d3=%d\n",
+					i, (unsigned)e1, (unsigned)e2, (int)d1, (int)d2, (int)d3);
+		}
+	}
+	ast_log(LOG_WARNING, "u-Law tandem transcoding test completed.\n");
+#endif /* TEST_TANDEM_TRANSCODING */
 }
 
